@@ -458,9 +458,7 @@ generate_fips186 (DSA_secret_key *sk, unsigned int nbits, unsigned int qbits,
                                              r_counter,
                                              r_seed, r_seedlen);
       else
-        ec = _gcry_generate_fips186_3_prime (nbits, qbits,
-                                             initial_seed.seed,
-                                             initial_seed.seedlen,
+        ec = _gcry_generate_fips186_3_prime (nbits, qbits, NULL, 0,
                                              &prime_q, &prime_p,
                                              r_counter,
                                              r_seed, r_seedlen, NULL);
@@ -637,16 +635,16 @@ sign (gcry_mpi_t r, gcry_mpi_t s, gcry_mpi_t input, DSA_secret_key *skey,
       k = _gcry_dsa_gen_k (skey->q, GCRY_STRONG_RANDOM);
     }
 
-  /* kinv = k^(-1) mod q */
-  kinv = mpi_alloc( mpi_get_nlimbs(k) );
-  mpi_invm(kinv, k, skey->q );
-
   _gcry_dsa_modify_k (k, skey->q, qbits);
 
   /* r = (a^k mod p) mod q */
   mpi_powm( r, skey->g, k, skey->p );
   mpi_fdiv_r( r, r, skey->q );
 
+  /* kinv = k^(-1) mod q */
+  kinv = mpi_alloc( mpi_get_nlimbs(k) );
+  mpi_invm(kinv, k, skey->q );
+
   /* s = (kinv * ( hash + x * r)) mod q */
   tmp = mpi_alloc( mpi_get_nlimbs(skey->p) );
   mpi_mul( tmp, skey->x, r );
