@@ -43,6 +43,7 @@ _gcry_ecc_curve_free (elliptic_curve_t *E)
   mpi_free (E->b);  E->b = NULL;
   _gcry_mpi_point_free_parts (&E->G);
   mpi_free (E->n);  E->n = NULL;
+  mpi_free (E->h);  E->h = NULL;
 }
 
 
@@ -60,10 +61,10 @@ _gcry_ecc_curve_copy (elliptic_curve_t E)
   R.p = mpi_copy (E.p);
   R.a = mpi_copy (E.a);
   R.b = mpi_copy (E.b);
-  _gcry_mpi_point_init (&R.G, 0);
+  _gcry_mpi_point_init (&R.G);
   point_set (&R.G, &E.G);
   R.n = mpi_copy (E.n);
-  R.h = E.h;
+  R.h = mpi_copy (E.h);
 
   return R;
 }
@@ -97,23 +98,19 @@ _gcry_ecc_dialect2str (enum ecc_dialects dialect)
     {
     case ECC_DIALECT_STANDARD:  str = "Standard"; break;
     case ECC_DIALECT_ED25519:   str = "Ed25519"; break;
-    case ECC_DIALECT_SAFECURVE: str = "SafeCurve"; break;
     }
   return str;
 }
 
 
-/* Return an uncompressed point (X,Y) in P as a malloced buffer with
- * its byte length stored at R_LENGTH.  May not be used for sensitive
- * data. */
-unsigned char *
-_gcry_ecc_ec2os_buf (gcry_mpi_t x, gcry_mpi_t y, gcry_mpi_t p,
-                     unsigned int *r_length)
+gcry_mpi_t
+_gcry_ecc_ec2os (gcry_mpi_t x, gcry_mpi_t y, gcry_mpi_t p)
 {
   gpg_err_code_t rc;
   int pbytes = (mpi_get_nbits (p)+7)/8;
   size_t n;
   unsigned char *buf, *ptr;
+  gcry_mpi_t result;
 
   buf = xmalloc ( 1 + 2*pbytes );
   *buf = 04; /* Uncompressed point.  */
@@ -136,35 +133,29 @@ _gcry_ecc_ec2os_buf (gcry_mpi_t x, gcry_mpi_t y, gcry_mpi_t p,
       memset (ptr, 0, (pbytes-n));
     }
 
-  *r_length = 1 + 2*pbytes;
-  return buf;
+  rc = _gcry_mpi_scan (&result, GCRYMPI_FMT_USG, buf, 1+2*pbytes, NULL);
+  if (rc)
+    log_fatal ("mpi_scan failed: %s\n", gpg_strerror (rc));
+  xfree (buf);
+
+  return result;
 }
 
 
-gcry_mpi_t
-_gcry_ecc_ec2os (gcry_mpi_t x, gcry_mpi_t y, gcry_mpi_t p)
-{
-  unsigned char *buf;
-  unsigned int buflen;
-
-  buf = _gcry_ecc_ec2os_buf (x, y, p, &buflen);
-  return mpi_set_opaque (NULL, buf, 8*buflen);
-}
-
 /* Convert POINT into affine coordinates using the context CTX and
    return a newly allocated MPI.  If the conversion is not possible
    NULL is returned.  This function won't print an error message.  */
 gcry_mpi_t
-_gcry_mpi_ec_ec2os (gcry_mpi_point_t point, mpi_ec_t ec)
+_gcry_mpi_ec_ec2os (gcry_mpi_point_t point, mpi_ec_t ectx)
 {
   gcry_mpi_t g_x, g_y, result;
 
-  g_x = mpi_new (ec->nbits);
-  g_y = mpi_new (ec->nbits);
-  if (_gcry_mpi_ec_get_affine (g_x, g_y, point, ec))
+  g_x = mpi_new (0);
+  g_y = mpi_new (0);
+  if (_gcry_mpi_ec_get_affine (g_x, g_y, point, ectx))
     result = NULL;
   else
-    result = _gcry_ecc_ec2os (g_x, g_y, ec->p);
+    result = _gcry_ecc_ec2os (g_x, g_y, ectx->p);
   mpi_free (g_x);
   mpi_free (g_y);
 
@@ -172,13 +163,12 @@ _gcry_mpi_ec_ec2os (gcry_mpi_point_t point, mpi_ec_t ec)
 }
 
 
-/* Decode octet string in VALUE into RESULT, in the format defined by SEC 1.
-   RESULT must have been initialized and is set on success to the
+/* RESULT must have been initialized and is set on success to the
    point given by VALUE.  */
-gpg_err_code_t
-_gcry_ecc_sec_decodepoint (gcry_mpi_t value, mpi_ec_t ec, mpi_point_t result)
+gcry_err_code_t
+_gcry_ecc_os2ec (mpi_point_t result, gcry_mpi_t value)
 {
-  gpg_err_code_t rc;
+  gcry_err_code_t rc;
   size_t n;
   const unsigned char *buf;
   unsigned char *buf_memory;
@@ -212,89 +202,30 @@ _gcry_ecc_sec_decodepoint (gcry_mpi_t value, mpi_ec_t ec, mpi_point_t result)
       xfree (buf_memory);
       return GPG_ERR_INV_OBJ;
     }
-
-  if (*buf == 2 || *buf == 3)
+  if (*buf != 4)
     {
-      gcry_mpi_t x3;
-      gcry_mpi_t t;
-      gcry_mpi_t p1_4;
-      int y_bit = (*buf == 3);
-
-      if (ec == NULL || !mpi_test_bit (ec->p, 1))
-        {
-          xfree (buf_memory);
-          return GPG_ERR_NOT_IMPLEMENTED; /* No support for point compression.  */
-        }
-
-      n = n - 1;
-      rc = _gcry_mpi_scan (&x, GCRYMPI_FMT_USG, buf+1, n, NULL);
       xfree (buf_memory);
-      if (rc)
-        return rc;
-
-      /*
-       * Recover Y.  The Weierstrass curve: y^2 = x^3 + a*x + b
-       */
-
-      x3 = mpi_new (ec->nbits);
-      t = mpi_new (ec->nbits);
-      p1_4 = mpi_new (ec->nbits);
-      y = mpi_new (ec->nbits);
-
-      /* Compute right hand side.  */
-      mpi_powm (x3, x, mpi_const (MPI_C_THREE), ec->p);
-      mpi_mul (t, ec->a, x);
-      mpi_mod (t, t, ec->p);
-      mpi_add (t, t, ec->b);
-      mpi_mod (t, t, ec->p);
-      mpi_add (t, t, x3);
-      mpi_mod (t, t, ec->p);
-
-      /*
-       * When p mod 4 = 3, modular square root of A can be computed by
-       * A^((p+1)/4) mod p
-       */
-
-      /* Compute (p+1)/4 into p1_4 */
-      mpi_rshift (p1_4, ec->p, 2);
-      _gcry_mpi_add_ui (p1_4, p1_4, 1);
-
-      mpi_powm (y, t, p1_4, ec->p);
-
-      if (y_bit != mpi_test_bit (y, 0))
-        mpi_sub (y, ec->p, y);
-
-      mpi_free (p1_4);
-      mpi_free (t);
-      mpi_free (x3);
+      return GPG_ERR_NOT_IMPLEMENTED; /* No support for point compression.  */
     }
-  else if (*buf == 4)
-    {
-      if ( ((n-1)%2) )
-        {
-          xfree (buf_memory);
-          return GPG_ERR_INV_OBJ;
-        }
-      n = (n-1)/2;
-      rc = _gcry_mpi_scan (&x, GCRYMPI_FMT_USG, buf+1, n, NULL);
-      if (rc)
-        {
-          xfree (buf_memory);
-          return rc;
-        }
-      rc = _gcry_mpi_scan (&y, GCRYMPI_FMT_USG, buf+1+n, n, NULL);
-      xfree (buf_memory);
-      if (rc)
-        {
-          mpi_free (x);
-          return rc;
-        }
-    }
-  else
+  if ( ((n-1)%2) )
     {
       xfree (buf_memory);
       return GPG_ERR_INV_OBJ;
     }
+  n = (n-1)/2;
+  rc = _gcry_mpi_scan (&x, GCRYMPI_FMT_USG, buf+1, n, NULL);
+  if (rc)
+    {
+      xfree (buf_memory);
+      return rc;
+    }
+  rc = _gcry_mpi_scan (&y, GCRYMPI_FMT_USG, buf+1+n, n, NULL);
+  xfree (buf_memory);
+  if (rc)
+    {
+      mpi_free (x);
+      return rc;
+    }
 
   mpi_set (result->x, x);
   mpi_set (result->y, y);
@@ -313,48 +244,45 @@ _gcry_ecc_sec_decodepoint (gcry_mpi_t value, mpi_ec_t ec, mpi_point_t result)
    is returned.  If G or D are given they override the values taken
    from EC. */
 mpi_point_t
-_gcry_ecc_compute_public (mpi_point_t Q, mpi_ec_t ec)
+_gcry_ecc_compute_public (mpi_point_t Q, mpi_ec_t ec,
+                          mpi_point_t G, gcry_mpi_t d)
 {
-  if (!ec->d || !ec->G || !ec->p || !ec->a)
+  if (!G)
+    G = ec->G;
+  if (!d)
+    d = ec->d;
+
+  if (!d || !G || !ec->p || !ec->a)
     return NULL;
   if (ec->model == MPI_EC_EDWARDS && !ec->b)
     return NULL;
 
-  if ((ec->dialect == ECC_DIALECT_ED25519 && (ec->flags & PUBKEY_FLAG_EDDSA))
-      || (ec->model == MPI_EC_EDWARDS && ec->dialect == ECC_DIALECT_SAFECURVE))
+  if (ec->dialect == ECC_DIALECT_ED25519
+      && (ec->flags & PUBKEY_FLAG_EDDSA))
     {
       gcry_mpi_t a;
       unsigned char *digest;
-      int b;
 
-      b = (ec->nbits+7)/8;
-      if (ec->nbits == 255)
-        ;
-      else if (ec->nbits == 448)
-        b++;
-      else
-        return NULL;            /* Not implemented.  */
-
-      if (_gcry_ecc_eddsa_compute_h_d (&digest, ec))
+      if (_gcry_ecc_eddsa_compute_h_d (&digest, d, ec))
         return NULL;
 
       a = mpi_snew (0);
-      _gcry_mpi_set_buffer (a, digest, b, 0);
+      _gcry_mpi_set_buffer (a, digest, 32, 0);
       xfree (digest);
 
       /* And finally the public key.  */
       if (!Q)
-        Q = mpi_point_new (ec->nbits);
+        Q = mpi_point_new (0);
       if (Q)
-        _gcry_mpi_ec_mul_point (Q, a, ec->G, ec);
+        _gcry_mpi_ec_mul_point (Q, a, G, ec);
       mpi_free (a);
     }
   else
     {
       if (!Q)
-        Q = mpi_point_new (ec->nbits);
+        Q = mpi_point_new (0);
       if (Q)
-        _gcry_mpi_ec_mul_point (Q, ec->d, ec->G, ec);
+        _gcry_mpi_ec_mul_point (Q, d, G, ec);
     }
 
   return Q;
@@ -362,54 +290,11 @@ _gcry_ecc_compute_public (mpi_point_t Q, mpi_ec_t ec)
 
 
 gpg_err_code_t
-_gcry_ecc_mont_encodepoint (gcry_mpi_t x, unsigned int nbits,
-                            int with_prefix,
-                            unsigned char **r_buffer, unsigned int *r_buflen)
+_gcry_ecc_mont_decodepoint (gcry_mpi_t pk, mpi_ec_t ctx, mpi_point_t result)
 {
   unsigned char *rawmpi;
   unsigned int rawmpilen;
 
-  rawmpi = _gcry_mpi_get_buffer_extra (x, (nbits+7)/8,
-                                       with_prefix? -1 : 0, &rawmpilen, NULL);
-  if (rawmpi == NULL)
-    return gpg_err_code_from_syserror ();
-
-  if (with_prefix)
-    {
-      rawmpi[0] = 0x40;
-      rawmpilen++;
-    }
-
-  *r_buffer = rawmpi;
-  *r_buflen = rawmpilen;
-  return 0;
-}
-
-
-gpg_err_code_t
-_gcry_ecc_mont_decodepoint (gcry_mpi_t pk, mpi_ec_t ec, mpi_point_t result)
-{
-  unsigned char *rawmpi;
-  unsigned int rawmpilen;
-  unsigned int nbytes = (ec->nbits+7)/8;
-
-  /*
-   * It is not reliable to assume that the first byte of 0x40
-   * means the prefix.
-   *
-   * For newer implementation, it is reliable since we always put
-   * 0x40 for x-only coordinate.
-   *
-   * For data by older implementation (non-released development
-   * version in 2015), there is no 0x40 prefix added.
-   *
-   * So, it is possible to have shorter length of data when it was
-   * handled as MPI, removing preceding zeros.
-   *
-   * Besides, when data was parsed as MPI, we might have 0x00
-   * prefix (when the MSB in the first byte is set).
-   */
-
   if (mpi_is_opaque (pk))
     {
       const unsigned char *buf;
@@ -420,36 +305,52 @@ _gcry_ecc_mont_decodepoint (gcry_mpi_t pk, mpi_ec_t ec, mpi_point_t result)
         return GPG_ERR_INV_OBJ;
       rawmpilen = (rawmpilen + 7)/8;
 
-      if (rawmpilen == nbytes + 1
-          && (buf[0] == 0x00 || buf[0] == 0x40))
+      if (rawmpilen > 1 && (rawmpilen%2) && buf[0] == 0x40)
         {
           rawmpilen--;
           buf++;
         }
-      else if (rawmpilen > nbytes)
-        return GPG_ERR_INV_OBJ;
 
-      rawmpi = xtrymalloc (nbytes);
+      rawmpi = xtrymalloc (rawmpilen? rawmpilen:1);
       if (!rawmpi)
         return gpg_err_code_from_syserror ();
 
       p = rawmpi + rawmpilen;
       while (p > rawmpi)
         *--p = *buf++;
-
-      if (rawmpilen < nbytes)
-        memset (rawmpi + nbytes - rawmpilen, 0, nbytes - rawmpilen);
     }
   else
     {
+      unsigned int nbytes = (ctx->nbits+7)/8;
+
       rawmpi = _gcry_mpi_get_buffer (pk, nbytes, &rawmpilen, NULL);
       if (!rawmpi)
         return gpg_err_code_from_syserror ();
-      if (rawmpilen > nbytes + BYTES_PER_MPI_LIMB)
-        {
-          xfree (rawmpi);
-          return GPG_ERR_INV_OBJ;
+      /*
+       * It is not reliable to assume that 0x40 means the prefix.
+       *
+       * For newer implementation, it is reliable since we always put
+       * 0x40 for x-only coordinate.
+       *
+       * For data with older implementation (non-released development
+       * version), it is possible to have the 0x40 as a part of data.
+       * Besides, when data was parsed as MPI, we might have 0x00
+       * prefix.
+       *
+       * So, we need to check if it's really the prefix or not.
+       * Only when it's the prefix, we remove it.
+       */
+      if (pk->nlimbs * BYTES_PER_MPI_LIMB < nbytes)
+        {/*
+          * It is possible for data created by older implementation
+          * to have shorter length when it was parsed as MPI.
+          */
+          unsigned int len = pk->nlimbs * BYTES_PER_MPI_LIMB;
+
+          memmove (rawmpi + nbytes - len, rawmpi, len);
+          memset (rawmpi, 0, nbytes - len);
         }
+
       /*
        * When we have the prefix (0x40 or 0x00), it comes at the end,
        * since it is taken by _gcry_mpi_get_buffer with little endian.
@@ -459,8 +360,7 @@ _gcry_ecc_mont_decodepoint (gcry_mpi_t pk, mpi_ec_t ec, mpi_point_t result)
       rawmpilen = nbytes;
     }
 
-  if ((ec->nbits % 8))
-    rawmpi[0] &= (1 << (ec->nbits % 8)) - 1;
+  rawmpi[0] &= (1 << (ctx->nbits % 8)) - 1;
   _gcry_mpi_set_buffer (result->x, rawmpi, rawmpilen, 0);
   xfree (rawmpi);
   mpi_set_ui (result->z, 1);
