@@ -25,8 +25,8 @@
  *                                    |
  *                                   \ /
  *                   global.c: [MM entrance points]   -----> [user callbacks]
- *                               |          |    
- *                               |          |    
+ *                               |          |
+ *                               |          |
  *                              \ /        \ /
  *
  *      stdmem.c: [non-secure handlers] [secure handlers]
@@ -49,6 +49,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <stdarg.h>
+#include <errno.h>
 
 #include "g10lib.h"
 #include "stdmem.h"
@@ -87,13 +88,17 @@ _gcry_private_enable_m_guard (void)
 void *
 _gcry_private_malloc (size_t n)
 {
-  if (!n) 
-    return NULL; /* Allocating 0 bytes is undefined - we better return
-                    an error to detect such coding errors.  */
-  if (use_m_guard) 
+  if (!n)
+    {
+      gpg_err_set_errno (EINVAL);
+      return NULL; /* Allocating 0 bytes is undefined - we better return
+                      an error to detect such coding errors.  */
+    }
+
+  if (use_m_guard)
     {
       char *p;
-      
+
       if ( !(p = malloc (n + EXTRA_ALIGN+5)) )
         return NULL;
       ((byte*)p)[EXTRA_ALIGN+0] = n;
@@ -103,7 +108,7 @@ _gcry_private_malloc (size_t n)
       p[4+EXTRA_ALIGN+n] = MAGIC_END_BYTE;
       return p+EXTRA_ALIGN+4;
     }
-  else 
+  else
     {
       return malloc( n );
     }
@@ -112,19 +117,24 @@ _gcry_private_malloc (size_t n)
 
 /*
  * Allocate memory of size N from the secure memory pool.  Return NULL
- * if we are out of memory.
+ * if we are out of memory.  XHINT tells the allocator that the caller
+ * used an xmalloc style call.
  */
 void *
-_gcry_private_malloc_secure (size_t n)
+_gcry_private_malloc_secure (size_t n, int xhint)
 {
-  if (!n) 
-    return NULL; /* Allocating 0 bytes is undefined - better return an
-                    error to detect such coding errors.  */
-  if (use_m_guard) 
+  if (!n)
+    {
+      gpg_err_set_errno (EINVAL);
+      return NULL; /* Allocating 0 bytes is undefined - better return an
+                      error to detect such coding errors.  */
+    }
+
+  if (use_m_guard)
     {
       char *p;
-      
-      if ( !(p = _gcry_secmem_malloc (n +EXTRA_ALIGN+ 5)) )
+
+      if (!(p = _gcry_secmem_malloc (n + EXTRA_ALIGN + 5, xhint)))
         return NULL;
       ((byte*)p)[EXTRA_ALIGN+0] = n;
       ((byte*)p)[EXTRA_ALIGN+1] = n >> 8 ;
@@ -135,27 +145,28 @@ _gcry_private_malloc_secure (size_t n)
     }
   else
     {
-      return _gcry_secmem_malloc( n );
+      return _gcry_secmem_malloc (n, xhint);
     }
 }
 
 
 /*
- * Realloc and clear the old space
- * Return NULL if there is not enough memory.
+ * Realloc and clear the old space.  XHINT tells the allocator that
+ * the caller used an xmalloc style call.  Returns NULL if there is
+ * not enough memory.
  */
 void *
-_gcry_private_realloc ( void *a, size_t n )
+_gcry_private_realloc (void *a, size_t n, int xhint)
 {
   if (use_m_guard)
     {
       unsigned char *p = a;
       char *b;
       size_t len;
-      
+
       if (!a)
         return _gcry_private_malloc(n);
-        
+
       _gcry_private_check_heap(p);
       len  = p[-4];
       len |= p[-3] << 8;
@@ -163,7 +174,7 @@ _gcry_private_realloc ( void *a, size_t n )
       if( len >= n ) /* We don't shrink for now. */
         return a;
       if (p[-1] == MAGIC_SEC_BYTE)
-        b = _gcry_private_malloc_secure(n);
+        b = _gcry_private_malloc_secure (n, xhint);
       else
         b = _gcry_private_malloc(n);
       if (!b)
@@ -175,7 +186,7 @@ _gcry_private_realloc ( void *a, size_t n )
     }
   else if ( _gcry_private_is_secure(a) )
     {
-      return _gcry_secmem_realloc( a, n );
+      return _gcry_secmem_realloc (a, n, xhint);
     }
   else
     {
@@ -191,10 +202,10 @@ _gcry_private_check_heap (const void *a)
     {
       const byte *p = a;
       size_t len;
-      
+
       if (!p)
         return;
-      
+
       if ( !(p[-1] == MAGIC_NOR_BYTE || p[-1] == MAGIC_SEC_BYTE) )
         _gcry_log_fatal ("memory at %p corrupted (underflow=%02x)\n", p, p[-1]);
       len  = p[-4];
@@ -219,17 +230,13 @@ _gcry_private_free (void *a)
   if (use_m_guard )
     {
       _gcry_private_check_heap(p);
-      if ( _gcry_private_is_secure(a) )
-        _gcry_secmem_free(p-EXTRA_ALIGN-4);
-      else
+      if (! _gcry_secmem_free (p - EXTRA_ALIGN - 4))
         {
-          free(p-EXTRA_ALIGN-4);
+          free (p - EXTRA_ALIGN - 4);
 	}
     }
-  else if ( _gcry_private_is_secure(a) )
-    _gcry_secmem_free(p);
-  else
-    free(p);
+  else if (!_gcry_secmem_free (p))
+    {
+      free(p);
+    }
 }
-
-
