@@ -27,7 +27,6 @@
 #include <assert.h>
 #include <errno.h>
 #include <unistd.h>
-#include <time.h>
 #if HAVE_PTHREAD
 # include <pthread.h>
 #endif
@@ -90,7 +89,6 @@ struct thread_arg_s
 
 
 
-#if defined(HAVE_PTHREAD) || defined(_WIN32)
 /* Wrapper functions to access Libgcrypt's internal test lock.  */
 static void
 external_lock_test_init (int line)
@@ -132,11 +130,10 @@ external_lock_test_destroy (int line)
   if (err)
     fail ("destroying lock failed at %d: %s", line, gpg_strerror (err));
 }
-#endif
+
 
 
 
-#if defined(HAVE_PTHREAD) || defined(_WIN32)
 /* The nonce thread.  We simply request a couple of nonces and
    return.  */
 static THREAD_RET_TYPE
@@ -156,7 +153,6 @@ nonce_thread (void *argarg)
   gcry_free (arg);
   return THREAD_RET_VALUE;
 }
-#endif
 
 
 /* To check our locking function we run several threads all accessing
@@ -212,8 +208,7 @@ check_nonce_lock (void)
       else
         info ("nonce thread %d has terminated", i);
     }
-#else
-  (void)arg;
+
 #endif /*!_WIN32*/
 }
 
@@ -253,7 +248,6 @@ print_accounts (void)
 }
 
 
-#if defined(HAVE_PTHREAD) || defined(_WIN32)
 /* Get a a random integer value in the range 0 to HIGH.  */
 static unsigned int
 get_rand (int high)
@@ -317,7 +311,6 @@ accountant_thread (void *arg)
     }
   return THREAD_RET_VALUE;
 }
-#endif
 
 
 static void
@@ -362,8 +355,7 @@ run_test (void)
     fail ("waiting for revision thread failed: %d", (int)GetLastError ());
   CloseHandle (rthread);
 
-  external_lock_test_destroy (__LINE__);
-#elif HAVE_PTHREAD
+#else /*!_WIN32*/
   pthread_t rthread;
   pthread_t athreads[N_ACCOUNTANTS];
   int rc, i;
@@ -392,8 +384,9 @@ run_test (void)
   else
     info ("revision thread has terminated");
 
-  external_lock_test_destroy (__LINE__);
 #endif /*!_WIN32*/
+
+  external_lock_test_destroy (__LINE__);
 }
 
 
@@ -436,15 +429,15 @@ main (int argc, char **argv)
   srand (time(NULL)*getpid());
 
   if (debug)
-    xgcry_control ((GCRYCTL_SET_DEBUG_FLAGS, 1u, 0));
-  xgcry_control ((GCRYCTL_DISABLE_SECMEM, 0));
+    xgcry_control (GCRYCTL_SET_DEBUG_FLAGS, 1u, 0);
+  xgcry_control (GCRYCTL_DISABLE_SECMEM, 0);
   if (!gcry_check_version (GCRYPT_VERSION))
     die ("version mismatch");
   /* We are using non-public interfaces - check the exact version.  */
   if (strcmp (gcry_check_version (NULL), GCRYPT_VERSION))
     die ("exact version match failed");
-  xgcry_control ((GCRYCTL_ENABLE_QUICK_RANDOM, 0));
-  xgcry_control ((GCRYCTL_INITIALIZATION_FINISHED, 0));
+  xgcry_control (GCRYCTL_ENABLE_QUICK_RANDOM, 0);
+  xgcry_control (GCRYCTL_INITIALIZATION_FINISHED, 0);
 
   check_nonce_lock ();
 
